Object.defineProperty(exports, '__esModule', { value: true });

var fs = require('fs/promises');
var path = require('path');
var rollup = require('rollup');
var fs$1 = require('fs');
var module$1 = require('module');
var rollupPluginSwc3 = require('rollup-plugin-swc3');
var commonjs = require('@rollup/plugin-commonjs');
var shebang = require('rollup-plugin-preserve-shebang');
var json = require('@rollup/plugin-json');
var pluginNodeResolve = require('@rollup/plugin-node-resolve');
var replace = require('@rollup/plugin-replace');

const rootDir = process.cwd();
var config = {
    rootDir
};

function getTypings(pkg) {
    return pkg.types || pkg.typings;
}
function getDistPath(distPath) {
    return path.resolve(config.rootDir, distPath);
}
function findExport(field) {
    if (!field) return;
    if (typeof field === 'string') return field;
    const value = field['.'] || field['import'] || field['module'] || field['default'];
    return findExport(value);
}
function parseExport(exportsCondition) {
    const paths = {};
    if (typeof exportsCondition === 'string') {
        paths.export = exportsCondition;
    } else {
        paths.main = paths.main || exportsCondition['require'] || exportsCondition['node'] || exportsCondition['default'];
        paths.module = paths.module || exportsCondition['module'];
        paths.export = findExport(exportsCondition);
    }
    return paths;
}
function getExportPaths(pkg) {
    const pathsMap = {};
    const mainExport = {};
    if (pkg.main) {
        mainExport.main = pkg.main;
    }
    if (pkg.module) {
        mainExport.module = pkg.module;
    }
    pathsMap['.'] = mainExport;
    const { exports: exportsConditions  } = pkg;
    if (exportsConditions) {
        if (typeof exportsConditions === 'string') {
            mainExport.export = exportsConditions;
        } else {
            const exportKeys = Object.keys(exportsConditions);
            if (exportKeys.some((key)=>key.startsWith('.'))) {
                exportKeys.forEach((subExport)=>{
                    pathsMap[subExport] = parseExport(exportsConditions[subExport]);
                });
            } else {
                Object.assign(mainExport, parseExport(exportsConditions));
            }
        }
    }
    pathsMap['.'] = mainExport;
    return pathsMap;
}
function getExportDist(pkg) {
    const paths = getExportPaths(pkg)['.'];
    const dist = [];
    if (paths.main) {
        dist.push({
            format: 'cjs',
            file: getDistPath(paths.main)
        });
    }
    if (paths.module) {
        dist.push({
            format: 'esm',
            file: getDistPath(paths.module)
        });
    }
    if (paths.export) {
        dist.push({
            format: 'esm',
            file: getDistPath(paths.export)
        });
    }
    // default fallback to output `dist/index.js` in default esm format
    if (dist.length === 0) {
        dist.push({
            format: 'esm',
            file: getDistPath('dist/index.js')
        });
    }
    return dist;
}
function getExportConditionDist(pkg, exportCondition) {
    // const pkgExports = pkg.exports || {}
    const dist = [];
    // "exports": "..."
    if (typeof exportCondition === 'string') {
        dist.push({
            format: pkg.type === 'module' ? 'esm' : 'cjs',
            file: getDistPath(exportCondition)
        });
    } else {
        // "./<subexport>": { }
        const subExports = exportCondition // pkgExports[subExport]
        ;
        // Ignore json exports, like "./package.json"
        // if (subExport.endsWith('.json')) return dist
        if (typeof subExports === 'string') {
            dist.push({
                format: 'esm',
                file: getDistPath(subExports)
            });
        } else {
            if (subExports.require) {
                dist.push({
                    format: 'cjs',
                    file: getDistPath(subExports.require)
                });
            }
            if (subExports.import) {
                dist.push({
                    format: 'esm',
                    file: getDistPath(subExports.import)
                });
            }
        }
    }
    return dist;
}

function asyncGeneratorStep$1(gen, resolve, reject, _next, _throw, key, arg) {
    try {
        var info = gen[key](arg);
        var value = info.value;
    } catch (error) {
        reject(error);
        return;
    }
    if (info.done) {
        resolve(value);
    } else {
        Promise.resolve(value).then(_next, _throw);
    }
}
function _async_to_generator$1(fn) {
    return function() {
        var self = this, args = arguments;
        return new Promise(function(resolve, reject) {
            var gen = fn.apply(self, args);
            function _next(value) {
                asyncGeneratorStep$1(gen, resolve, reject, _next, _throw, "next", value);
            }
            function _throw(err) {
                asyncGeneratorStep$1(gen, resolve, reject, _next, _throw, "throw", err);
            }
            _next(undefined);
        });
    };
}
function exit(err) {
    logger.error(err);
    process.exit(1);
}
const formatDuration = (duration)=>duration >= 1000 ? `${duration / 1000}s` : `${duration}ms`;
function getPackageMeta(cwd) {
    return _getPackageMeta.apply(this, arguments);
}
function _getPackageMeta() {
    _getPackageMeta = _async_to_generator$1(function*(cwd) {
        const pkgFilePath = path.resolve(cwd, 'package.json');
        let targetPackageJson = {};
        try {
            targetPackageJson = JSON.parse((yield fs.readFile(pkgFilePath, {
                encoding: 'utf-8'
            })));
        } catch (_) {}
        return targetPackageJson;
    });
    return _getPackageMeta.apply(this, arguments);
}
const logger = {
    log (arg) {
        console.log(arg);
    },
    warn (arg) {
        console.log('\x1b[33m' + arg + '\x1b[0m');
    },
    error (arg) {
        console.error('\x1b[31m' + arg + '\x1b[0m');
    }
};
function isTypescript(filename) {
    const ext = path.extname(filename);
    return ext === '.ts' || ext === '.tsx';
}
function fileExists(filePath) {
    return _fileExists.apply(this, arguments);
}
function _fileExists() {
    _fileExists = _async_to_generator$1(function*(filePath) {
        try {
            yield fs.access(filePath);
            return true;
        } catch (err) {
            if (err.code === 'ENOENT') {
                return false;
            }
            throw err;
        }
    });
    return _fileExists.apply(this, arguments);
}
const isNotNull = (n)=>Boolean(n);

function _extends$1() {
    _extends$1 = Object.assign || function(target) {
        for(var i = 1; i < arguments.length; i++){
            var source = arguments[i];
            for(var key in source){
                if (Object.prototype.hasOwnProperty.call(source, key)) {
                    target[key] = source[key];
                }
            }
        }
        return target;
    };
    return _extends$1.apply(this, arguments);
}
const minifyOptions = {
    compress: true,
    format: {
        comments: 'some',
        wrapFuncArgs: false,
        preserveAnnotations: true
    },
    mangle: {
        toplevel: true
    }
};
let hasLoggedTsWarning = false;
function resolveTypescript(cwd) {
    let ts;
    const m = new module$1.Module('', undefined);
    m.paths = module$1.Module._nodeModulePaths(cwd);
    try {
        ts = m.require('typescript');
    } catch (_) {
        console.error(_);
        if (!hasLoggedTsWarning) {
            hasLoggedTsWarning = true;
            exit('Could not load TypeScript compiler. Try to install `typescript` as dev dependency');
        }
    }
    return ts;
}
function getBuildEnv(envs) {
    if (!envs.includes('NODE_ENV')) {
        envs.push('NODE_ENV');
    }
    const envVars = envs.reduce((acc, key)=>{
        const value = process.env[key];
        if (typeof value !== 'undefined') {
            acc['process.env.' + key] = JSON.stringify(value);
        }
        return acc;
    }, {});
    return envVars;
}
function buildInputConfig(entry, pkg, options, { tsConfigPath , tsCompilerOptions , dtsOnly  }) {
    var _options_external;
    const externals = options.noExternal ? [] : [
        pkg.peerDependencies,
        pkg.dependencies,
        pkg.peerDependenciesMeta
    ].filter((n)=>Boolean(n)).map((o)=>Object.keys(o)).reduce((a, b)=>a.concat(b), []).concat(((_options_external = options.external) != null ? _options_external : []).concat(pkg.name ? [
        pkg.name
    ] : []));
    const { useTypescript , runtime , target: jscTarget , minify  } = options;
    const hasSpecifiedTsTarget = Boolean((tsCompilerOptions == null ? void 0 : tsCompilerOptions.target) && tsConfigPath);
    const plugins = (dtsOnly ? [
        shebang(),
        useTypescript && require('rollup-plugin-dts').default({
            compilerOptions: _extends$1({}, tsCompilerOptions, {
                declaration: true,
                noEmit: false,
                noEmitOnError: true,
                emitDeclarationOnly: true,
                checkJs: false,
                declarationMap: false,
                skipLibCheck: true,
                preserveSymlinks: false,
                target: 'esnext',
                module: 'esnext',
                incremental: false,
                jsx: tsCompilerOptions.jsx || 'react'
            })
        })
    ] : [
        replace({
            values: getBuildEnv(options.env || []),
            preventAssignment: true
        }),
        pluginNodeResolve.nodeResolve({
            preferBuiltins: runtime === 'node',
            extensions: [
                '.mjs',
                '.js',
                '.json',
                '.node',
                '.jsx'
            ]
        }),
        commonjs({
            include: /node_modules\//
        }),
        json(),
        shebang(),
        rollupPluginSwc3.swc({
            include: /\.(m|c)?[jt]sx?$/,
            exclude: 'node_modules',
            tsconfig: tsConfigPath,
            jsc: _extends$1({}, !hasSpecifiedTsTarget && {
                target: jscTarget
            }, {
                loose: true,
                externalHelpers: false,
                parser: {
                    syntax: useTypescript ? 'typescript' : 'ecmascript',
                    [useTypescript ? 'tsx' : 'jsx']: true,
                    privateMethod: true,
                    classPrivateProperty: true,
                    exportDefaultFrom: true
                }
            }, minify && {
                minify: _extends$1({}, minifyOptions, {
                    sourceMap: options.sourcemap
                })
            }),
            sourceMaps: options.sourcemap,
            inlineSourcesContent: false
        })
    ]).filter(isNotNull);
    return {
        input: entry,
        external (id) {
            return externals.some((name)=>id === name || id.startsWith(name + '/'));
        },
        plugins,
        treeshake: {
            propertyReadSideEffects: false
        },
        onwarn (warning, warn) {
            const code = warning.code || '';
            // Some may not have types, like CLI binary
            if (dtsOnly && code === 'EMPTY_BUNDLE') return;
            if ([
                'MIXED_EXPORTS',
                'PREFER_NAMED_EXPORTS',
                'UNRESOLVED_IMPORT',
                'THIS_IS_UNDEFINED'
            ].includes(code)) return;
            // If the circular dependency warning is from node_modules, ignore it
            if (code === 'CIRCULAR_DEPENDENCY' && /Circular dependency: node_modules/.test(warning.message)) {
                return;
            }
            warn(warning);
        }
    };
}
function buildOutputConfigs(options, pkg, { tsCompilerOptions , dtsOnly  }) {
    const { format , exportCondition  } = options;
    const exportPaths = getExportPaths(pkg);
    // respect if tsconfig.json has `esModuleInterop` config;
    // add ESModule mark if cjs and ESModule are both generated;
    // TODO: support `import` in exportCondition
    const mainExport = exportPaths['.'];
    const useEsModuleMark = Boolean(tsCompilerOptions.esModuleInterop || mainExport.main && mainExport.module);
    const typings = getTypings(pkg);
    const file = options.file && path.resolve(config.rootDir, options.file);
    const dtsDir = typings ? path.dirname(path.resolve(config.rootDir, typings)) : path.resolve(config.rootDir, 'dist');
    // file base name without extension
    const name = file ? file.replace(new RegExp(`${path.extname(file)}$`), '') : undefined;
    const dtsFile = file ? name + '.d.ts' : (exportCondition == null ? void 0 : exportCondition.name) ? path.resolve(dtsDir, (exportCondition.name === '.' ? 'index' : exportCondition.name) + '.d.ts') : typings && path.resolve(config.rootDir, typings);
    // If there's dts file, use `output.file`
    const dtsPathConfig = dtsFile ? {
        file: dtsFile
    } : {
        dir: dtsDir
    };
    return _extends$1({
        name: pkg.name || name
    }, dtsOnly ? dtsPathConfig : {
        file: file
    }, {
        format,
        exports: 'named',
        esModule: useEsModuleMark,
        freeze: false,
        strict: false,
        sourcemap: options.sourcemap
    });
}
function buildConfig(entry, pkg, bundleConfig, dtsOnly) {
    var _options_exportCondition;
    const { file  } = bundleConfig;
    const useTypescript = isTypescript(entry);
    const options = _extends$1({}, bundleConfig, {
        useTypescript
    });
    let tsCompilerOptions = {};
    let tsConfigPath;
    if (useTypescript) {
        const ts = resolveTypescript(config.rootDir);
        tsConfigPath = path.resolve(config.rootDir, 'tsconfig.json');
        if (fs$1.existsSync(tsConfigPath)) {
            const basePath = tsConfigPath ? path.dirname(tsConfigPath) : config.rootDir;
            const tsconfigJSON = ts.readConfigFile(tsConfigPath, ts.sys.readFile).config;
            tsCompilerOptions = ts.parseJsonConfigFileContent(tsconfigJSON, ts.sys, basePath).options;
        } else {
            tsConfigPath = undefined;
            exit('tsconfig.json is missing in your project directory');
        }
    }
    const typescriptOptions = {
        dtsOnly,
        tsConfigPath,
        tsCompilerOptions
    };
    const inputOptions = buildInputConfig(entry, pkg, options, typescriptOptions);
    const outputExports = options.exportCondition ? getExportConditionDist(pkg, options.exportCondition.export) : getExportDist(pkg);
    let outputConfigs = [];
    // Generate dts job - single config
    if (dtsOnly) {
        outputConfigs = [
            buildOutputConfigs(_extends$1({}, bundleConfig, {
                format: 'es',
                useTypescript
            }), pkg, typescriptOptions)
        ];
    } else {
        // multi outputs with specified format
        outputConfigs = outputExports.map((exportDist)=>{
            return buildOutputConfigs(_extends$1({}, bundleConfig, {
                file: exportDist.file,
                format: exportDist.format,
                useTypescript
            }), pkg, typescriptOptions);
        });
        // CLI output option is always prioritized
        if (file) {
            var _outputExports_;
            const fallbackFormat = (_outputExports_ = outputExports[0]) == null ? void 0 : _outputExports_.format;
            outputConfigs = [
                buildOutputConfigs(_extends$1({}, bundleConfig, {
                    file,
                    format: bundleConfig.format || fallbackFormat,
                    useTypescript
                }), pkg, typescriptOptions)
            ];
        }
    }
    return {
        input: inputOptions,
        output: outputConfigs,
        exportName: ((_options_exportCondition = options.exportCondition) == null ? void 0 : _options_exportCondition.name) || '.',
        dtsOnly
    };
}

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
    try {
        var info = gen[key](arg);
        var value = info.value;
    } catch (error) {
        reject(error);
        return;
    }
    if (info.done) {
        resolve(value);
    } else {
        Promise.resolve(value).then(_next, _throw);
    }
}
function _async_to_generator(fn) {
    return function() {
        var self = this, args = arguments;
        return new Promise(function(resolve, reject) {
            var gen = fn.apply(self, args);
            function _next(value) {
                asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
            }
            function _throw(err) {
                asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
            }
            _next(undefined);
        });
    };
}
function _extends() {
    _extends = Object.assign || function(target) {
        for(var i = 1; i < arguments.length; i++){
            var source = arguments[i];
            for(var key in source){
                if (Object.prototype.hasOwnProperty.call(source, key)) {
                    target[key] = source[key];
                }
            }
        }
        return target;
    };
    return _extends.apply(this, arguments);
}
function _object_without_properties_loose(source, excluded) {
    if (source == null) return {};
    var target = {};
    var sourceKeys = Object.keys(source);
    var key, i;
    for(i = 0; i < sourceKeys.length; i++){
        key = sourceKeys[i];
        if (excluded.indexOf(key) >= 0) continue;
        target[key] = source[key];
    }
    return target;
}
const SRC = 'src' // resolve from src/ directory
;
function logBuild(exportPath, dtsOnly, duration) {
    logger.log(` ✓  ${dtsOnly ? 'Typed' : 'Built'} ${exportPath} ${formatDuration(duration)}`);
}
function assignDefault(options, name, defaultValue) {
    if (!(name in options) || options[name] == null) {
        options[name] = defaultValue;
    }
}
function resolveSourceFile(cwd, filename) {
    return path.resolve(cwd, SRC, filename);
}
function getSourcePathFromExportPath(cwd, exportPath) {
    return _getSourcePathFromExportPath.apply(this, arguments);
}
function _getSourcePathFromExportPath() {
    _getSourcePathFromExportPath = // Map '.' -> './index.[ext]'
    // Map './lite' -> './lite.[ext]'
    // Return undefined if no match or if it's package.json exports
    _async_to_generator(function*(cwd, exportPath) {
        const exts = [
            'js',
            'cjs',
            'mjs',
            'jsx',
            'ts',
            'tsx'
        ];
        for (const ext of exts){
            // ignore package.json
            if (exportPath.endsWith('package.json')) return;
            if (exportPath === '.') exportPath = './index';
            const filename = resolveSourceFile(cwd, `${exportPath}.${ext}`);
            if (yield fileExists(filename)) {
                return filename;
            }
        }
        return;
    });
    return _getSourcePathFromExportPath.apply(this, arguments);
}
function hasMultiEntryExport(pkg) {
    const packageExportsField = pkg.exports || {};
    if (typeof packageExportsField === 'string') return false;
    const exportKeys = (packageExportsField ? Object.keys(packageExportsField) : []).filter((key)=>key !== './package.json');
    return exportKeys.length > 0 && exportKeys.every((name)=>name.startsWith('.'));
}
function bundle(entryPath) {
    return _bundle.apply(this, arguments);
}
function _bundle() {
    _bundle = _async_to_generator(function*(entryPath, _param = {}) {
        var { cwd  } = _param, options = _object_without_properties_loose(_param, [
            "cwd"
        ]);
        config.rootDir = path.resolve(process.cwd(), cwd || '');
        assignDefault(options, 'format', 'es');
        assignDefault(options, 'minify', false);
        assignDefault(options, 'target', 'es2015');
        const pkg = yield getPackageMeta(config.rootDir);
        const packageExportsField = pkg.exports || {};
        const isMultiEntries = hasMultiEntryExport(pkg);
        // Handle single entry file
        if (!isMultiEntries) {
            // Use specified string file path if possible, then fallback to the default behavior entry picking logic
            // e.g. "exports": "./dist/index.js" -> use "./index.<ext>" as entry
            entryPath = entryPath || (yield getSourcePathFromExportPath(config.rootDir, '.')) || '';
        }
        function buildEntryConfig(packageExports, dtsOnly) {
            return _buildEntryConfig.apply(this, arguments);
        }
        function _buildEntryConfig() {
            _buildEntryConfig = _async_to_generator(function*(packageExports, dtsOnly) {
                const configs = Object.keys(packageExports).map(/*#__PURE__*/ _async_to_generator(function*(entryExport) {
                    const source = yield getSourcePathFromExportPath(config.rootDir, entryExport);
                    if (!source) return undefined;
                    if (dtsOnly && !isTypescript(source)) return;
                    options.exportCondition = {
                        source,
                        name: entryExport,
                        export: packageExports[entryExport]
                    };
                    const entry = resolveSourceFile(cwd, source);
                    const rollupConfig = buildConfig(entry, pkg, options, dtsOnly);
                    return rollupConfig;
                }));
                return (yield Promise.all(configs)).filter((n)=>Boolean(n));
            });
            return _buildEntryConfig.apply(this, arguments);
        }
        const bundleOrWatch = (rollupConfig)=>{
            const { input , exportName  } = rollupConfig;
            const exportPath = getExportPath(pkg, exportName);
            // Log original entry file relative path
            const source = typeof input.input === 'string' ? path.relative(config.rootDir, input.input) : exportPath;
            const buildMetadata = {
                source
            };
            if (options.watch) {
                return Promise.resolve(runWatch(rollupConfig, buildMetadata));
            }
            return runBundle(rollupConfig, buildMetadata);
        };
        const hasSpecifiedEntryFile = entryPath ? (yield fileExists(entryPath)) && (yield fs.stat(entryPath)).isFile() : false;
        if (!hasSpecifiedEntryFile && !isMultiEntries) {
            const err = new Error(`Entry file \`${entryPath}\` is not existed`);
            err.name = 'NOT_EXISTED';
            return Promise.reject(err);
        }
        // has `types` field in package.json or has `types` exports in any export condition for multi-entries
        const hasTypings = !!getTypings(pkg) || typeof packageExportsField === 'object' && Array.from(Object.values(packageExportsField || {})).some((condition)=>condition.hasOwnProperty('types'));
        // Enable types generation if it's types field specified in package.json
        if (hasTypings) {
            options.dts = hasTypings;
        }
        if (isMultiEntries) {
            const pkgExports = packageExportsField;
            const buildConfigs = yield buildEntryConfig(pkgExports, false);
            const assetsJobs = buildConfigs.map((rollupConfig)=>bundleOrWatch(rollupConfig));
            const typesJobs = options.dts ? (yield buildEntryConfig(pkgExports, true)).map((rollupConfig)=>bundleOrWatch(rollupConfig)) : [];
            return yield Promise.all(assetsJobs.concat(typesJobs));
        }
        // Generate types
        if (isTypescript(entryPath) && options.dts) {
            yield bundleOrWatch(buildConfig(entryPath, pkg, options, true));
        }
        const rollupConfig = buildConfig(entryPath, pkg, options, false);
        return bundleOrWatch(rollupConfig);
    });
    return _bundle.apply(this, arguments);
}
// . -> pkg name
// ./lite -> <pkg name>/lite
function getExportPath(pkg, exportName) {
    const name = pkg.name || path.basename(config.rootDir);
    if (exportName === '.' || !exportName) return name;
    return path.join(name, exportName);
}
function runWatch({ input , output , dtsOnly  }, metadata) {
    const watchOptions = [
        _extends({}, input, {
            output: output,
            watch: {
                exclude: [
                    'node_modules/**'
                ]
            }
        })
    ];
    const watcher = rollup.watch(watchOptions);
    let startTime = Date.now();
    watcher.on('event', (event)=>{
        switch(event.code){
            case 'ERROR':
                {
                    return onError(event.error);
                }
            case 'START':
                {
                    startTime = Date.now();
                    logger.log(`Start building ${metadata.source} ...`);
                    break;
                }
            case 'END':
                {
                    const duration = Date.now() - startTime;
                    logBuild(metadata.source, dtsOnly, duration);
                    break;
                }
            default:
                return;
        }
    });
    return watcher;
}
function runBundle({ input , output , dtsOnly  }, jobOptions) {
    const startTime = Date.now();
    return rollup.rollup(input).then((bundle)=>{
        const writeJobs = output.map((options)=>bundle.write(options));
        return Promise.all(writeJobs);
    }, onError).then(()=>{
        const duration = Date.now() - startTime;
        logBuild(jobOptions.source, dtsOnly, duration);
    });
}
function onError(error) {
    if (!error) return;
    // logging source code in format
    if (error.frame) {
        process.stderr.write(error.frame + '\n');
    }
    // filter out the rollup plugin error information such as loc/frame/code...
    const err = new Error(error.message);
    err.stack = error.stack;
    throw err;
}

exports.bundle = bundle;
